// ========================================== //
//                                            //
//                   //
//               GZip                 //
//             2019-03-06               //
//                                            //
//                HiAsm                 //
//              www.hiasm.com                 //
//                                            //
//             : Netspirit               //
//         E-mail: netspirit@meta.ua          //
//                                            //
// ========================================== //


unit hiGZIP;


interface

uses
  Windows, Kol, Share, KolZLib;


type
  
  
  ThiGZIP = class
    private
      FAborted, FProcessing: Boolean;
      function CalcHeaderLength(Buffer: Pointer; BufLen: Integer): Integer;
      
    public
      ErrorCode: Integer;
      
      _prop_Level: Integer;
      _prop_InBuffer: Integer;
      _prop_OutBuffer: Integer;
      _prop_InclHeader: Integer;
      _prop_InclTrailer: Integer;
      _prop_DstFileName: string;
      _prop_SrcFileName: string;
      
      _data_Count: THI_Event;
      _data_Data: THI_Event;
      _data_DstFileName: THI_Event;
      _data_DstStream: THI_Event;
      _data_SrcFileName: THI_Event;
      _data_SrcStream: THI_Event;
      
      _event_onError: THI_Event;
      _event_onResult: THI_Event;
      
      constructor Create;

      function CompressString(var SrcStr: string): Boolean;
      function DecompressString(var SrcStr: string): Boolean;
      
      function CompressStream(SrcStream, DstStream: PStream; var Count: LongWord): Boolean;
      function DecompressStream(SrcStream, DstStream: PStream; var Count: LongWord): Boolean;
      
      procedure _work_doAbort(var _Data: TData; Index: Word);
      procedure _work_doCompress(var _Data: TData; Index: Word);
      procedure _work_doDecompress(var _Data: TData; Index: Word);
      procedure _work_doCompressFile(var _Data: TData; Index: Word);  
      procedure _work_doDecompressFile(var _Data: TData; Index: Word);
      procedure _work_doCompressStream(var _Data: TData; Index: Word);  
      procedure _work_doDecompressStream(var _Data: TData; Index: Word);
      
      procedure _work_doLevel(var _Data: TData; Index: Word); 
  end;


const

  ERROR_INCORRECT_DATA = 0;
  ERROR_PROCESSING = 1;
  ERROR_CANNOT_OPEN_SRC_FILE = 2;
  ERROR_CANNOT_OPEN_DST_FILE = 3;
  ERROR_STREAM_READ = 4;
  ERROR_STREAM_WRITE = 5;
  ERROR_UNKNOWN = 6;

  DefaultHeader: TBaseGZipHeader = 
  (
    Sig: GZIP_SIGNATURE;
    CM: Z_DEFLATED;
    Flags: 0;
    MTime: 0;
    ExtraFlags: 0;
    OS: 0;
  );
  
  TRAILER_SIZE = 8;


implementation


//  Position            (    ). 
//    ,      2  (  Integer).
//     ,      4 .
procedure SetStreamPosition(st: PStream; NewPos: Integer);
begin
  if NewPos < 0 then
  begin
    st.Seek($7FFFFFFF, spBegin); //   2   
    st.Seek(Cardinal(NewPos)-$7FFFFFFF, spCurrent); //        
  end
  else
    st.Seek(NewPos, spBegin);
end;




constructor ThiGZIP.Create;
begin
  inherited;
  _prop_InBuffer := 512*1024;
  _prop_OutBuffer := 512*1024;
  _prop_Level := 8;
  _prop_InclHeader := 1;
  _prop_InclTrailer := 1;
end;

function ThiGZIP.CalcHeaderLength(Buffer: Pointer; BufLen: Integer): Integer; // 0 - 
var
  L, K: Integer;
  F: Byte;
begin
  Result := 0;
  
  if (BufLen < SizeOf(TBaseGZipHeader)) or
     (TBaseGZipHeader(Buffer^).Sig <> GZIP_SIGNATURE)
  then Exit;
  
  F := TBaseGZipHeader(Buffer^).Flags;
  L := SizeOf(TBaseGZipHeader);
  
  Dec(BufLen, L);
  Buffer := Pointer(Integer(Buffer) + L);
  
  //FEXTRA - 2   +   
  if F and FEXTRA <> 0 then
  begin
    if BufLen < 2 then Exit;
    K := PWord(Buffer)^;
    Inc(K, 2);
    if K > BufLen then Exit;
    
    L := L + K;
    Dec(BufLen, K);
    Buffer := Pointer(Integer(Buffer) + K);
  end;
  
  //FNAME -     0
  if F and FNAME <> 0 then
  begin
    if BufLen < 1 then Exit;
    K := StrLen(Buffer) + 1;
    if K > BufLen then Exit;

    L := L + K;
    Dec(BufLen, K);
    Buffer := Pointer(Integer(Buffer) + K);
  end;
  
  //FCOMMENT -    0
  if F and FCOMMENT <> 0 then
  begin
    if BufLen < 1 then Exit;
    K := StrLen(Buffer) + 1;
    if K > BufLen then Exit;

    L := L + K;
    Dec(BufLen, K);
    //Buffer := Pointer(Integer(Buffer) + K);
  end;
  
  //FHCRC - 2   
  if F and FHCRC <> 0 then
  begin
    if BufLen < 2 then Exit;
    Inc(L, 2);
  end;
  
  Result := L;
end;

function ThiGZIP.CompressString(var SrcStr: string): Boolean;
var 
  ZRec: TZStreamRec;
  OutBuf: string;
  err: Integer;
  OrigL, ResL, P: Integer;
label
  finish;
begin
  Result := False;
  ErrorCode := ERROR_UNKNOWN;
  
  OrigL := Length(SrcStr);
  ResL := OrigL + Trunc(OrigL*0.2 + 12); //       ,   (  )
  
  if _prop_InclHeader = 1 then Inc(ResL, SizeOf(TBaseGZipHeader));
  if _prop_InclTrailer = 1 then Inc(ResL, TRAILER_SIZE);
  
  SetLength(OutBuf, ResL);

  FillChar(ZRec, SizeOf(TZStreamRec), 0);
  
  if _prop_InclHeader = 1 then
  begin
    PBaseGZipHeader(@OutBuf[1])^ := DefaultHeader;
    ZRec.avail_out := ResL - SizeOf(TBaseGZipHeader);
    P := SizeOf(TBaseGZipHeader) + 1;
  end
  else
  begin
    ZRec.avail_out := ResL;
    P := 1;
  end;
  ZRec.next_out := @OutBuf[P];

  
  ZRec.next_in := @SrcStr[1];
  ZRec.avail_in := OrigL;
 
  if deflateInit2_(ZRec, _prop_Level, Z_DEFLATED, -15, 8, Z_DEFAULT_STRATEGY, zlib_version, SizeOf(TZStreamRec)) <> Z_OK then Exit;
  
  
  while True do
  begin

    //         (      )
    if ZRec.avail_out = 0 then
    begin
      P := Length(OutBuf);
      SetLength(OutBuf, P + 1000);
      ZRec.next_out := @OutBuf[P+1]; //      
      ZRec.avail_out := 1000;
    end;
    
    err := deflate(ZRec, Z_FINISH);
    
    if err = Z_STREAM_END then //  
    begin
      if _prop_InclHeader = 1 then Inc(ZRec.total_out, SizeOf(TBaseGZipHeader));
      
      //    ,  
      if _prop_InclTrailer = 1 then
      begin
        if ZRec.avail_out <> TRAILER_SIZE then SetLength(OutBuf, Length(OutBuf) - ZRec.avail_out + TRAILER_SIZE);
        PInteger(@OutBuf[ZRec.total_out + 1])^ := crc32(0, @SrcStr[1], OrigL);
        PInteger(@OutBuf[ZRec.total_out + 5])^ := ZRec.total_in;
      end
      else
        if ZRec.avail_out <> 0 then SetLength(OutBuf, Length(OutBuf) - ZRec.avail_out);
        
      SrcStr := OutBuf;
      Break;
    end
    else
      if err <> Z_OK then //   
      begin
        ErrorCode := ERROR_PROCESSING;
        goto finish;
      end;      

  end;

  Result := True;
  
finish:
  deflateEnd(ZRec);
end;

function ThiGZIP.DecompressString(var SrcStr: string): Boolean;
var 
  ZRec: TZStreamRec;
  OutBuf: string;
  err: Integer;
  OrigL, Avail, P: Integer;
begin
  Result := False;
  ErrorCode := ERROR_UNKNOWN;

  OrigL := Length(SrcStr);
  
  if _prop_InclTrailer = 1 then
  begin
    if OrigL <= TRAILER_SIZE then
    begin
      ErrorCode := ERROR_INCORRECT_DATA;
      Exit;
    end
    else
      Dec(OrigL, TRAILER_SIZE);
  end;

  
  FillChar(ZRec, SizeOf(TZStreamRec), 0);
  
  //    
  if _prop_InclHeader = 1 then
  begin
    err := CalcHeaderLength(@SrcStr[1], OrigL);
    if err = 0 then
    begin
      ErrorCode := ERROR_INCORRECT_DATA;
      Exit;
    end;
    Dec(OrigL, err);
    ZRec.next_in := @SrcStr[err+1];
  end
  else
    ZRec.next_in := @SrcStr[1];

  //         
  // (         )
  {if OrigL = 0 then
  begin
    SrcStr := '';
    Result := True;
    Exit;
  end;}
  
  ZRec.avail_in := OrigL;
  
  if inflateInit2_(ZRec, -15, zlib_version, SizeOf(TZStreamRec)) <> Z_OK then Exit;
  
  Avail := OrigL * 3; //    -  3   
  
  while True do
  begin
  
    //if ZRec.avail_out = 0 then //  0,   -   ZRec.avail_out = 0
    //begin
      //           
      P := Length(OutBuf); //      
      SetLength(OutBuf, P + Avail);
      ZRec.next_out := @OutBuf[P+1];
      ZRec.avail_out := Avail;
    //end;
    
    err := inflate(ZRec, Z_NO_FLUSH);
    
    // Z_OK:
    // 1)  ZRec.avail_out = 0    
    // 2)  Z_NO_FLUSH     
    
    // Z_BUF_ERROR:
    // 1)   
    // 2)      Z_FINISH
    
    // Z_DATA_ERROR:  
    
    if (err = Z_STREAM_END) or (err = Z_OK) then //   
    begin
      //  Z_OK  ZRec.avail_out = 0     .
      if (Err = Z_OK) and (ZRec.avail_out = 0) then Continue;
      
      //  ,  Z_OK    ZRec.avail_out > 0
      //   -   Z_STREAM_END -  () ?
      //   ,   .
      
      //        -     
      if ZRec.avail_out > 0 then SetLength(OutBuf, Length(OutBuf) - ZRec.avail_out);
      
      SrcStr := OutBuf;
      Result := True;
    end
    else
    begin
      ErrorCode := ERROR_PROCESSING; 
    end;
    
    Break;
  end;
  
  
  inflateEnd(ZRec);
end;


function ThiGZIP.CompressStream(SrcStream, DstStream: PStream; var Count: LongWord): Boolean;
var 
  ZRec: TZStreamRec;
  InBuf, OutBuf: array of Byte;
  err: Integer;
  Need, R: LongWord;
  Flush: Integer;
label
  finish;
begin
  Result := False;
  ErrorCode := ERROR_UNKNOWN;

  
  FillChar(ZRec, SizeOf(TZStreamRec), 0);
  
  if deflateInit2_(ZRec, _prop_Level, Z_DEFLATED, -15, 8, Z_DEFAULT_STRATEGY, zlib_version, SizeOf(TZStreamRec)) <> Z_OK then Exit;
  
  
  if _prop_InclHeader = 1 then
  begin
    ZRec.total_out := DstStream.Write(DefaultHeader, SizeOf(TBaseGZipHeader));
    if ZRec.total_out < SizeOf(TBaseGZipHeader) then
    begin
      ErrorCode := ERROR_STREAM_WRITE;
      goto finish;
    end;
  end;

  SetLength(InBuf, _prop_InBuffer); // TODO: SetLength(InBuf, Min(_prop_InBuffer, Count))
  SetLength(OutBuf, _prop_OutBuffer); 
  
  ZRec.adler := 0; //  deflateInit2_() ZRec.adler=1 (init  Adler32).    CRC32  init=0
  
  Flush := Z_NO_FLUSH;
  
  repeat
    R := 0;
    
    if Count > 0 then
    begin
      if Count < LongWord(_prop_InBuffer) then
        Need := Count
      else
        Need := LongWord(_prop_InBuffer);
      
      R := SrcStream.Read(InBuf[0], Need);
      
      if R = 0 then
      begin
        ErrorCode := ERROR_STREAM_READ;
        goto finish;
      end;
      
      Dec(Count, R);
    end;
    
    if Count = 0 then Flush := Z_FINISH;
    
    //  
    if FAborted then goto finish;

    ZRec.avail_in := R;
    ZRec.next_in := @InBuf[0];
    ZRec.adler := crc32(ZRec.adler, @InBuf[0], R);
    
    
    repeat
      ZRec.next_out := @OutBuf[0];
      ZRec.avail_out := _prop_OutBuffer;
      
      err := deflate(ZRec, Flush); // no bad return value
      
      if (err <> Z_OK) and (err <> Z_STREAM_END) then
      begin
        ErrorCode := ERROR_PROCESSING;
        goto finish;
      end;
      
      //   
      Need := _prop_OutBuffer - ZRec.avail_out;
      if Need > 0 then
      begin
        R := DstStream.Write(OutBuf[0], Need);
        if R < Need then
        begin
          ErrorCode := ERROR_STREAM_WRITE;
          goto finish;
        end;
      end;
      
      if FAborted then goto finish;
      
    until (ZRec.avail_out <> 0);
    
  until (Count = 0);
  


  //  gzip- -      (<> mod 2^32)
  if _prop_InclTrailer = 1 then
  begin
    PInteger(@OutBuf[0])^ := ZRec.adler;
    PInteger(@OutBuf[4])^ := ZRec.total_in;
    Inc(ZRec.total_out, DstStream.Write(OutBuf[0], TRAILER_SIZE));
  end;
  
  Result := True;
  
finish:
  deflateEnd(ZRec);
  Count := ZRec.total_out;
end;


function ThiGZIP.DecompressStream(SrcStream, DstStream: PStream; var Count: LongWord): Boolean;
var 
  ZRec: TZStreamRec;
  InBuf, OutBuf: array of Byte;
  err: Integer;
  Need, R: LongWord;
label
  finish;
begin
  Result := False;
  ErrorCode := ERROR_UNKNOWN;
  
  //   ,   
  if _prop_InclTrailer = 1 then
  begin
    if Count <= TRAILER_SIZE then
    begin
      ErrorCode := ERROR_INCORRECT_DATA;
      Exit;
    end
    else
      Dec(Count, TRAILER_SIZE);
  end;

  FillChar(ZRec, SizeOf(TZStreamRec), 0);
  
  if inflateInit2_(ZRec, -15, zlib_version, SizeOf(TZStreamRec)) <> Z_OK then Exit;
  
  SetLength(InBuf, _prop_InBuffer);
  SetLength(OutBuf, _prop_OutBuffer);
  
  while Count > 0 do
  begin
    if Count < LongWord(_prop_InBuffer) then
      Need := Count
    else
      Need := LongWord(_prop_InBuffer);
    
    R := SrcStream.Read(InBuf[0], Need);
    
    if R = 0 then
    begin
      ErrorCode := ERROR_STREAM_READ;
      goto finish;
    end;
    
    Dec(Count, R);
    
    if FAborted then goto finish;
    
    //       
    if (ZRec.total_in = 0) and (_prop_InclHeader = 1) then 
    begin
      err := CalcHeaderLength(@InBuf[0], R);
      if err = 0 then
      begin
        ErrorCode := ERROR_INCORRECT_DATA;
        goto finish;
      end;
      ZRec.next_in := @InBuf[err];
      ZRec.avail_in := R - LongWord(err);
    end
    else
    begin
      ZRec.next_in := @InBuf[0];
      ZRec.avail_in := R;
    end;
    
    //    ZRec.next_in,
    //   ZRec.next_out  .
    //   -   
    repeat
      ZRec.avail_out := _prop_OutBuffer;
      ZRec.next_out := @OutBuf[0];
      
      err := inflate(ZRec, Z_NO_FLUSH);
      
      //   
      if (err <> Z_OK) and (err <> Z_STREAM_END) then
      begin
        ErrorCode := ERROR_PROCESSING;
        goto finish;
      end;
      
      Need := _prop_OutBuffer - ZRec.avail_out;
      if Need > 0 then
      begin
        R := DstStream.Write(OutBuf[0], Need);
        if R < Need then
        begin
          ErrorCode := ERROR_STREAM_WRITE;
          goto finish;
        end;
      end;
      
      if FAborted then goto finish;
      
    until (ZRec.avail_out <> 0);
    //     err = Z_STREAM_END  ( Count),
    //     ,   
  end;
  
  Result := True;
  
  //   ,      -
  //      
  if _prop_InclTrailer = 1 then
    SrcStream.Seek(TRAILER_SIZE, spCurrent);
  
finish:  
  inflateEnd(ZRec);
  Count := ZRec.total_out;
end;


procedure ThiGZIP._work_doAbort(var _Data: TData; Index: Word);
begin
  FProcessing := False;
  FAborted := True;
end;


procedure ThiGZIP._work_doCompress(var _Data: TData; Index: Word);
var 
  S: string;
begin
  if FProcessing then Exit;
  FProcessing := True;
  
  S := ReadString(_Data, _data_Data, '');  

  if CompressString(S) then
    _hi_CreateEvent(_Data, @_event_onResult, S)
  else
    _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  
  FProcessing := False;
end;

procedure ThiGZIP._work_doDecompress(var _Data: TData; Index: Word);
var
  S: string;
begin
  if FProcessing then Exit;
  FProcessing := True;
  
  S := ReadString(_Data, _data_Data, '');  

  if DecompressString(S) then
    _hi_CreateEvent(_Data, @_event_onResult, S)
  else
    _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  
  FProcessing := False;
end;

procedure ThiGZIP._work_doCompressFile(var _Data: TData; Index: Word);
var 
  Src, Dst: PStream;
  C: LongWord;
  Fn: string;
label
  finish;
begin
  if FProcessing then Exit;
  FProcessing := True;
  FAborted := False;

  Dst := nil;
 
  Src := NewReadFileStream(ReadString(_Data, _data_SrcFileName, _prop_SrcFileName));
  if Src.Handle = INVALID_HANDLE_VALUE then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_SRC_FILE);
    goto finish;
  end;
  
  Fn := ReadString(_Data, _data_DstFileName, _prop_DstFileName);
  Dst := NewFileStream(Fn, ofOpenWrite or ofCreateAlways or ofShareDenyWrite);
  if Dst.Handle = INVALID_HANDLE_VALUE then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_DST_FILE);
    goto finish;
  end;
  
  C := Src.Size;
  
  if not CompressStream(Src, Dst, C) then
  begin
    Dst.Free;
    Dst := nil;
    DeleteFile(PChar(Fn));
    if not FAborted then
      _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  end
  else
    _hi_CreateEvent(_Data, @_event_onResult, Integer(C));
  
finish:
  src.Free;
  dst.Free;
  
  FProcessing := False;
end;


procedure ThiGZIP._work_doDecompressFile(var _Data: TData; Index: Word);
var 
  Src, Dst: PStream;
  C: LongWord;
  Fn: string;
label
  finish;
begin
  if FProcessing then Exit;
  FProcessing := True;
  FAborted := False;

  Dst := nil;
 
  Src := NewReadFileStream(ReadString(_Data, _data_SrcFileName, _prop_SrcFileName));
  if Src.Handle = INVALID_HANDLE_VALUE then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_SRC_FILE);
    goto finish;
  end;
  
  Fn := ReadString(_Data, _data_DstFileName, _prop_DstFileName);
  Dst := NewFileStream(Fn, ofOpenWrite or ofCreateAlways or ofShareDenyWrite);
  if Dst.Handle = INVALID_HANDLE_VALUE then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_DST_FILE);
    goto finish;
  end;
  
  C := Src.Size;
  
  if not DecompressStream(Src, Dst, C) then
  begin
    Dst.Free;
    Dst := nil;
    DeleteFile(PChar(Fn));
    if not FAborted then
      _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  end
  else
    _hi_CreateEvent(_Data, @_event_onResult, Integer(C));
  
finish:
  src.Free;
  dst.Free;
  
  FProcessing := False;
end;


procedure ThiGZIP._work_doCompressStream(var _Data: TData; Index: Word);
var 
  Src, Dst: PStream;
  C: LongWord;
  P1, P2: Integer;
begin
  if FProcessing then Exit;
  FProcessing := True;
  FAborted := False;
  
  Src := ReadStream(_Data, _data_SrcStream);
  Dst := ReadStream(_Data, _data_DstStream);
  C := ReadInteger(_Data, _data_Count, 0);
  
  if Src = nil then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_SRC_FILE);
    Exit;
  end;
  
  if Dst = nil then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_DST_FILE);
    Exit;
  end;
  
  //       
  P1 := Src.Position;
  P2 := Dst.Position;
  
  if not CompressStream(Src, Dst, C) then
  begin
    SetStreamPosition(Src, P1);
    SetStreamPosition(Dst, P2);
    if not FAborted then _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  end
  else
    _hi_CreateEvent(_Data, @_event_onResult, Integer(C));
    
  
  FProcessing := False;
end;

procedure ThiGZIP._work_doDecompressStream(var _Data: TData; Index: Word);
var 
  Src, Dst: PStream;
  C: LongWord;
  P1, P2: Integer;
begin
  if FProcessing then Exit;
  FProcessing := True;
  FAborted := False;
  
  Src := ReadStream(_Data, _data_SrcStream);
  Dst := ReadStream(_Data, _data_DstStream);
  C := ReadInteger(_Data, _data_Count, 0);
  
  if Src = nil then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_SRC_FILE);
    Exit;
  end;
  
  if Dst = nil then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_DST_FILE);
    Exit;
  end;
  
  //       
  P1 := Src.Position;
  P2 := Dst.Position;
  
  if not DecompressStream(Src, Dst, C) then
  begin
    SetStreamPosition(Src, P1);
    SetStreamPosition(Dst, P2);
    if not FAborted then _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  end
  else
    _hi_CreateEvent(_Data, @_event_onResult, Integer(C));
  
  FProcessing := False;
end;

procedure ThiGZIP._work_doLevel(var _Data: TData; Index: Word);
begin
  _prop_Level := ToInteger(_Data);
  if _prop_Level < 0 then _prop_Level := 0
  else
    if _prop_Level > 9 then _prop_Level := 9;
end;


end.